# Adapted from https://github.com/tqdm/tqdm#usage

from time import sleep
import contextlib
import sys
from tqdm import tqdm
from tqdm.contrib import DummyTqdmFile


@contextlib.contextmanager
def std_out_err_redirect_tqdm():
    orig_out_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = map(DummyTqdmFile, orig_out_err)
        yield orig_out_err[0]
    finally:
        sys.stdout, sys.stderr = orig_out_err


@contextlib.contextmanager
def tqdm_with_redirect(*args, **kwargs):
    with std_out_err_redirect_tqdm() as orig_stdout:
        with tqdm(file=orig_stdout, dynamic_ncols=True, *args, **kwargs) as t:
            yield t



from typing import Union
from enum import Enum
from .types import EnumList, VersionState

editable_version_states = [
    VersionState.PREPARE_FOR_SUBMISSION,
    VersionState.WAITING_FOR_REVIEW,
    VersionState.WAITING_FOR_EXPORT_COMPLIANCE,
    VersionState.REJECTED,
    VersionState.METADATA_REJECTED,
    VersionState.DEVELOPER_REJECTED,
]

live_version_state = VersionState.READY_FOR_SALE


def enum_name(x: Union[Enum, str]):  # pylint: disable=unsubscriptable-object
    return x.name if isinstance(x, Enum) else x


def enum_names(x_list: EnumList):
    return (enum_name(x) for x in x_list)


def version_state_is_editable(
    version_state: Union[VersionState, str]  # pylint: disable=unsubscriptable-object
) -> bool:
    """Test whether or not the version state is 'editable' in the App Store."""
    return enum_name(version_state) in enum_names(editable_version_states)


def version_state_is_live(
    version_state: Union[VersionState, str]  # pylint: disable=unsubscriptable-object
) -> bool:
    """Test whether or not the version state is 'live' in the App Store."""
    return enum_name(version_state) == enum_name(live_version_state)



import pygments
from pygments.lexers.data import JsonLexer
from pygments.formatters.terminal import TerminalFormatter
import json
import colorama
from typing import Any


def json_term(obj: Any):
    """Stringifies the object with json syntax highlighting for the terminal."""
    obj_formatted = json.dumps(obj, indent=2)
    return pygments.highlight(
        code=obj_formatted, lexer=JsonLexer(), formatter=TerminalFormatter()
    )


def clr(*colored_text: str, sep: str = "") -> str:
    """Add the color reset code after each colored_text."""
    return (sep + colorama.Style.RESET_ALL).join(
        colored_text
    ) + colorama.Style.RESET_ALL


def print_clr(*colored_text: str, sep: str = ""):
    """Add the color reset code after each colored_text."""
    print(clr(*colored_text, sep=sep))


EXTRA_INFO_COLOR = colorama.Style.DIM
KEYWORD_COLOR = colorama.Style.DIM + colorama.Fore.LIGHTCYAN_EX
USAGE_COLOR = colorama.Style.DIM + colorama.Fore.CYAN

clr_extra = lambda x: clr(EXTRA_INFO_COLOR + x)
clr_keyword = lambda x: clr(KEYWORD_COLOR + x)
clr_usage = lambda x: clr(USAGE_COLOR + x)



import logging
import colorama
from enum import Enum, auto
from typing import Union, Optional
from appstore_tools import appstore
from appstore_tools.print_util import print_clr, clr, json_term


class Verbosity(Enum):
    SHORT = auto()
    LONG = auto()
    FULL = auto()


def list_categories(
    access_token: str,
    platforms: appstore.PlatformList,
    verbosity: Verbosity = Verbosity.SHORT,
):
    """List the appstore's heirachry of categories and subcategories."""
    categories = appstore.get_categories(access_token=access_token, platforms=platforms)
    categories.sort(key=lambda x: x["id"])

    if verbosity == Verbosity.SHORT:
        for x in categories:
            print_clr(
                x["id"],
                colorama.Fore.LIGHTBLACK_EX
                + f' {{{",".join(x["attributes"]["platforms"])}}}',
            )

            for sub in x["relationships"]["subcategories"]["data"]:
                print_clr(colorama.Style.DIM + f'  {sub["id"]}')

    else:
        print(json_term(categories))


def list_apps(access_token: str, verbosity: Verbosity = Verbosity.SHORT):
    """List the apps found on the appstore."""
    apps = appstore.get_apps(access_token=access_token)
    if verbosity == Verbosity.SHORT:
        apps = [
            {
                "id": v["id"],
                "name": v["attributes"]["name"],
                "bundleId": v["attributes"]["bundleId"],
                "primaryLocale": v["attributes"]["primaryLocale"],
            }
            for v in apps
        ]
    elif verbosity == Verbosity.LONG:
        apps = [
            {
                "id": v["id"],
                **v["attributes"],
            }
            for v in apps
        ]
    print(json_term(apps))


def list_versions(
    access_token: str,
    app_id: str,
    platforms: appstore.PlatformList,
    states: appstore.VersionStateList,
    verbosity: Verbosity = Verbosity.SHORT,
):
    """List the app versions found on the appstore."""
    versions = appstore.get_versions(
        app_id=app_id, access_token=access_token, platforms=platforms, states=states
    )
    if verbosity == Verbosity.SHORT:
        versions = [
            {
                "id": x["id"],
                "platform": x["attributes"]["platform"],
                "versionString": x["attributes"]["versionString"],
                "appStoreState": x["attributes"]["appStoreState"],
            }
            for x in versions
        ]
    print(json_term({"appId": app_id, "versions": versions}))


def list_infos(
    access_token: str,
    app_id: str,
    states: appstore.VersionStateList,
    verbosity: Verbosity = Verbosity.SHORT,
):
    """List the app infos found on the appstore."""
    infos = appstore.get_infos(app_id=app_id, access_token=access_token, states=states)

    if verbosity == Verbosity.SHORT:
        infos = [
            {
                "id": x["id"],
                "appStoreState": x["attributes"]["appStoreState"],
                "appStoreAgeRating": x["attributes"]["appStoreAgeRating"],
                "brazilAgeRating": x["attributes"]["brazilAgeRating"],
                "kidsAgeBand": x["attributes"]["kidsAgeBand"],
            }
            for x in infos
        ]
    elif verbosity == Verbosity.LONG:
        infos_selected = []
        for x in infos:
            relationships = {}
            for k in x["relationships"]:
                relationships[k] = x["relationships"][k]["links"]["related"]
            infos_selected.append(
                {
                    "id": x["id"],
                    "appStoreState": x["attributes"]["appStoreState"],
                    "appStoreAgeRating": x["attributes"]["appStoreAgeRating"],
                    "brazilAgeRating": x["attributes"]["brazilAgeRating"],
                    "kidsAgeBand": x["attributes"]["kidsAgeBand"],
                    **relationships,
                }
            )
        infos = infos_selected

    print(json_term({"appId": app_id, "infos": infos}))


def list_screenshots(
    access_token: str,
    app_id: str,
    platforms: appstore.PlatformList,
    states: appstore.VersionStateList,
    version_limit: Optional[int],  # pylint: disable=unsubscriptable-object
    verbosity: Verbosity = Verbosity.SHORT,
):
    """List screenhots for each screenshot set of each app version."""
    logging.info(clr(colorama.Fore.GREEN + "app_id: ", str(app_id)))

    versions = appstore.get_versions(
        app_id=app_id, access_token=access_token, platforms=platforms, states=states
    )
    for version in versions[:version_limit]:
        version_id = version["id"]
        version_state = version["attributes"]["appStoreState"]
        print_clr(
            f"{colorama.Fore.GREEN}{version_state} {colorama.Style.DIM}{version_id} "
        )

        localizations = appstore.get_version_localizations(
            version_id=version_id, access_token=access_token
        )

        for loc in localizations:
            loc_id = loc["id"]
            locale = loc["attributes"]["locale"]

            screenshot_sets = appstore.get_screenshot_sets(
                localization_id=loc_id, access_token=access_token
            )
            print_clr(
                f"{colorama.Fore.GREEN}{locale}: ",
                f"Found {colorama.Fore.CYAN}{len(screenshot_sets)}{colorama.Fore.RESET} screenshot sets.",
            )

            for screenshot_set in screenshot_sets:
                ss_set_id = screenshot_set["id"]
                ss_display_type = screenshot_set["attributes"]["screenshotDisplayType"]
                screenshots = appstore.get_screenshots(
                    screenshot_set_id=ss_set_id, access_token=access_token
                )
                if verbosity == Verbosity.SHORT:
                    print_clr(colorama.Fore.CYAN + ss_display_type)
                    for x in screenshots:
                        print_clr(
                            f'  {colorama.Style.DIM}{x["attributes"]["assetDeliveryState"]["state"]:15}',
                            f'  {x["attributes"]["fileName"]}',
                        )
                    continue
                if verbosity == Verbosity.LONG:
                    screenshots = [
                        (
                            {
                                "id": x["id"],
                                "fileSize": x["attributes"]["fileSize"],
                                "fileName": x["attributes"]["fileName"],
                                "sourceFileChecksum": x["attributes"][
                                    "sourceFileChecksum"
                                ],
                                "templateUrl": x["attributes"]["imageAsset"][
                                    "templateUrl"
                                ],
                                "width": x["attributes"]["imageAsset"]["width"],
                                "height": x["attributes"]["imageAsset"]["height"],
                                "assetDeliveryState": x["attributes"][
                                    "assetDeliveryState"
                                ]["state"],
                            }
                            if x["attributes"]["assetDeliveryState"]["state"]
                            == appstore.MediaAssetState.COMPLETE
                            else {
                                "id": x["id"],
                                "fileSize": x["attributes"]["fileSize"],
                                "fileName": x["attributes"]["fileName"],
                                "sourceFileChecksum": x["attributes"][
                                    "sourceFileChecksum"
                                ],
                                "assetDeliveryState": x["attributes"][
                                    "assetDeliveryState"
                                ]["state"],
                            }
                        )
                        for x in screenshots
                    ]
                print(
                    json_term(
                        {
                            "screenshotDisplayType": ss_display_type,
                            "screenshots": screenshots,
                        }
                    )
                )


def list_previews(
    access_token: str,
    app_id: str,
    platforms: appstore.PlatformList,
    states: appstore.VersionStateList,
    version_limit: Optional[int],  # pylint: disable=unsubscriptable-object
):
    """List previews for each preview set of each app version."""
    logging.info(clr(colorama.Fore.GREEN + "app_id: ", str(app_id)))

    versions = appstore.get_versions(
        app_id=app_id, access_token=access_token, platforms=platforms, states=states
    )
    for version in versions[:version_limit]:
        version_id = version["id"]
        version_state = version["attributes"]["appStoreState"]
        print_clr(
            f"{colorama.Fore.GREEN}version: {colorama.Fore.BLUE}{version_id} {version_state}"
        )

        localizations = appstore.get_version_localizations(
            version_id=version_id, access_token=access_token
        )

        localization_ids = (l["id"] for l in localizations)
        for loc_id in localization_ids:
            preview_sets = appstore.get_preview_sets(
                localization_id=loc_id, access_token=access_token
            )
            print_clr(
                f"{colorama.Fore.GREEN}loc_id {loc_id}: ",
                f"Found {colorama.Fore.CYAN}{len(preview_sets)}{colorama.Fore.RESET} preview sets.",
            )

            for preview_set in preview_sets:
                preview_set_id = preview_set["id"]
                preview_type = preview_set["attributes"]["previewType"]
                preview_set = appstore.get_previews(
                    preview_set_id=preview_set_id, access_token=access_token
                )
                print_clr(f"{colorama.Fore.GREEN}previewType: ", preview_type)
                print(json_term(preview_set))



import configargparse
import argparse
import logging
import appstore_tools.console_actions as console_actions
import appstore_tools.appstore as appstore
import appstore_tools.actions as actions
import requests
import sys
from appstore_tools.print_util import print_clr, clr, clr_extra, clr_keyword, clr_usage
import colorama
import argparse_color_formatter
import appstore_tools.version as version

DEFAULT_CONFIG_FILES = ["appstore_tools.config"]
DEFAULT_ASSET_DIR = "appstore"

LOGO_ART = """                                                           
 _____         _____ _                  _____         _     
|  _  |___ ___|   __| |_ ___ ___ ___   |_   _|___ ___| |___ 
|     | . | . |__   |  _| . |  _| -_|    | | | . | . | |_ -|
|__|__|  _|  _|_____|_| |___|_| |___|    |_| |___|___|_|___|
      |_| |_|                                               
"""


class PrettyHelpFormatter(argparse_color_formatter.ColorRawDescriptionHelpFormatter):
    def _get_help_string(self, action):
        help = action.help
        if action.choices is not None:
            choice_strs = [str(choice) for choice in action.choices]
            choices_colored = [clr_keyword(choice) for choice in choice_strs]
            choices_joined = clr_extra(", ").join(choices_colored)
            help += clr_extra(" {") + choices_joined + clr_extra("}")

        if "%(default)" not in action.help:
            if (
                action.default is not configargparse.SUPPRESS
                and action.default is not None
            ):
                defaulting_nargs = [
                    configargparse.OPTIONAL,
                    configargparse.ZERO_OR_MORE,
                ]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += (
                        clr_extra(" (default:")
                        + clr_extra("%(default)s")
                        + clr_extra(")")
                    )
        return help

    def _format_usage(self, usage, actions, groups, prefix):
        if prefix is not None or usage is not None:
            return super()._format_usage(usage, actions, groups, prefix)

        prefix = f"{LOGO_ART}\nUsage:\n  "
        usage_colored = clr_usage(
            super()._format_usage(
                usage,
                actions,
                groups,
                "",
            )
        )
        return prefix + usage_colored

    def _format_action(self, action):
        action_invocation = super()._format_action_invocation(action)
        action_invocation_colored = clr_keyword(action_invocation)
        action_text = super()._format_action(action)
        return action_text.replace(action_invocation, action_invocation_colored, 1)


def arg_type_positive_int(arg):
    i = int(arg)
    if i < 1:
        raise configargparse.ArgumentTypeError(
            f"{arg} is an invalid positive int value"
        )
    return i


def add_config_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "-c",
        "--config-file",
        is_config_file=True,
        help="Args that start with '--' (eg. --log-level) can also be set in a config file (run.config or specified via -c). "
        + "Config file syntax allows: key=value, flag=true, stuff=[a,b,c] "
        + clr_extra("(details here https://goo.gl/R74nmi)")
        + ". "
        + "If an arg is specified in more than one place, then commandline values override config file values which override defaults.",
    )


def add_version_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--version",
        action="version",
        version=version.version,
    )


def add_help_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message."
        + clr_extra(" (help for actions: ")
        + clr_keyword("action --help")
        + clr_extra(")"),
    )


def add_log_level_argument(parser: configargparse.ArgumentParser):
    log_level_choices = list(logging._levelToName.values())
    parser.add_argument(
        "--log-level",
        choices=log_level_choices,
        default=logging.getLevelName(logging.WARNING),
        metavar="LOG_LEVEL",
        help="Set the logging level.",
    )


def add_platform_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--platform",
        choices=list(x.name for x in appstore.Platform),
        default=appstore.Platform.IOS.name,
        metavar="PLATFORM",
        help="Specify the target platform.",
    )


def add_platform_filter_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--platform",
        choices=list(x.name for x in appstore.Platform),
        metavar="PLATFORM",
        help="Filter by target platform.",
    )


def add_editable_filter_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--editable",
        action="store_true",
        help='Filter for versions in an "editable" state.'
        + clr_extra(
            " (details here https://help.apple.com/app-store-connect/#/dev18557d60e)"
        ),
    )


def add_live_filter_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--live", action="store_true", help='Filter for versions in an "live" state.'
    )


def add_version_state_filter_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--version-state",
        choices=list(x.name for x in appstore.VersionState),
        metavar="VERSION_STATE",
        help="Filter by appstore version state.",
    )


def add_info_filters_group(parser: configargparse.ArgumentParser):
    filter_group = parser.add_argument_group(
        title="Result filters",
    )
    add_editable_filter_argument(filter_group)
    add_live_filter_argument(filter_group)
    add_version_state_filter_argument(filter_group)


def add_version_filters_group(parser: configargparse.ArgumentParser):
    filter_group = parser.add_argument_group(
        title="Result filters",
    )
    add_platform_filter_argument(filter_group)
    add_editable_filter_argument(filter_group)
    add_live_filter_argument(filter_group)
    add_version_state_filter_argument(filter_group)


def create_platform_filter_list(args):
    return [args.platform] if args.platform is not None else list(appstore.Platform)


def create_version_state_filter_list(args):
    return (
        appstore.editable_version_states
        if args.editable
        else appstore.live_version_state
        if args.live
        else [args.version_state]
        if args.version_state is not None
        else list(appstore.VersionState)
    )


def add_verbosity_arguments(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--long",
        action="store_const",
        default=actions.Verbosity.SHORT,
        const=actions.Verbosity.LONG,
        dest="verbosity",
        help="Print out more information.",
    )
    parser.add_argument(
        "--full",
        action="store_const",
        const=actions.Verbosity.FULL,
        dest="verbosity",
        help="Print out full information.",
    )


def add_global_group(parser: configargparse.ArgumentParser):
    global_group = parser.add_argument_group(
        title="General",
    )
    add_help_argument(global_group)
    add_version_argument(global_group)
    add_config_argument(global_group)
    add_log_level_argument(global_group)


def add_asset_dir_argument(parser: configargparse.ArgumentParser):
    parser.add_argument(
        "--asset-dir",
        default=DEFAULT_ASSET_DIR,
        help="The directory where appstore assets are placed.",
    )


def add_subparser(subparsers: argparse._SubParsersAction, name: str, help: str):
    parser = subparsers.add_parser(
        name=name,
        help=help,
        default_config_files=DEFAULT_CONFIG_FILES,
        add_config_file_help=False,
        add_help=False,
        epilog=" ",  # leave an empty line after the message
        formatter_class=PrettyHelpFormatter,
    )
    add_global_group(parser)
    return parser


def add_authentication_group(parser: configargparse.ArgumentParser):
    auth_group = parser.add_argument_group(
        title="Authentication",
        description=clr_extra(
            "Authentication details are configured (and can be copied from) AppStore Connect->Users & Access->Keys."
        ),
    )
    auth_group.add_argument("--issuer-id", required=True, help="Issuer ID.")
    auth_group.add_argument("--key-id", required=True, help="Key ID.")
    key_group = auth_group.add_mutually_exclusive_group(required=True)
    key_group.add_argument("--key", help="Private Key as a string.")
    key_group.add_argument(
        "--key-file",
        type=configargparse.FileType(mode="r"),
        help="Private Key from a filepath.",
    )


def add_app_id_group(parser: configargparse.ArgumentParser):
    app_group = parser.add_argument_group(
        title="App ID",
        description=clr_extra(
            "App can either be identified by App ID (integer) or Bundle ID (string)."
        ),
    )
    key_group = app_group.add_mutually_exclusive_group(required=True)
    key_group.add_argument(
        "--app-id",
        type=int,
        help="The integer App ID assigned by the appstore.",
    )
    key_group.add_argument(
        "--bundle-id",
        help='The App\'s Bundle ID in the form "com.example.myapp".',
    )


def run():
    """Run appstore-tools from the command-line."""
    # Global
    global_parser = configargparse.ArgParser(
        default_config_files=DEFAULT_CONFIG_FILES,
        add_config_file_help=False,
        formatter_class=PrettyHelpFormatter,
        add_help=False,
        epilog=" ",  # leave an empty line after the message
    )
    add_global_group(global_parser)

    # Action subparsers
    action_subparsers = global_parser.add_subparsers(
        title="Action commands",
        dest="action",
        metavar="action",
        required=True,
        parser_class=configargparse.ArgParser,
    )

    # Action: apps
    apps_parser = add_subparser(
        action_subparsers,
        "apps",
        help="List all apps under the app store account.",
    )
    apps_group = apps_parser.add_argument_group(
        title="Apps",
    )
    add_verbosity_arguments(apps_group)
    add_authentication_group(apps_parser)

    # Action: infos
    infos_parser = add_subparser(
        action_subparsers,
        "infos",
        help="List all app infos.",
    )
    infos_group = infos_parser.add_argument_group(
        title="Infos",
    )
    add_verbosity_arguments(infos_group)
    add_info_filters_group(infos_parser)
    add_authentication_group(infos_parser)
    add_app_id_group(infos_parser)

    # Action: versions
    versions_parser = add_subparser(
        action_subparsers,
        "versions",
        help="List all app versions.",
    )
    versions_group = versions_parser.add_argument_group(
        title="Versions",
    )
    add_verbosity_arguments(versions_group)
    add_version_filters_group(versions_parser)
    add_authentication_group(versions_parser)
    add_app_id_group(versions_parser)

    # Action: screenshots
    screenshots_parser = add_subparser(
        action_subparsers,
        "screenshots",
        help="List the screenshots for an app.",
    )
    screenshots_group = screenshots_parser.add_argument_group(
        title="Screenshots",
    )
    add_verbosity_arguments(screenshots_group)
    screenshots_group.add_argument(
        "--version-limit",
        type=arg_type_positive_int,
        default=1,
        help="Limit the number of app versions displayed.",
    )
    add_version_filters_group(screenshots_parser)
    add_authentication_group(screenshots_parser)
    add_app_id_group(screenshots_parser)

    # Action: previews
    previews_parser = add_subparser(
        action_subparsers,
        "previews",
        help="List the previews for an app.",
    )
    previews_group = previews_parser.add_argument_group(
        title="Previews",
    )
    add_verbosity_arguments(previews_group)
    previews_group.add_argument(
        "--version-limit",
        type=arg_type_positive_int,
        default=1,
        help="Limit the number of app versions displayed.",
    )
    add_version_filters_group(previews_parser)
    add_authentication_group(previews_parser)
    add_app_id_group(previews_parser)

    # Action: categories
    categories_parser = add_subparser(
        action_subparsers,
        "categories",
        help="List the appstore's heirarchy of categories and subcategories.",
    )
    categories_group = categories_parser.add_argument_group(
        title="Categories",
    )
    add_platform_filter_argument(categories_group)
    add_verbosity_arguments(categories_group)
    add_authentication_group(categories_parser)

    # Action: download
    download_parser = add_subparser(
        action_subparsers,
        "download",
        help="Download all assets for an app.",
    )
    download_group = download_parser.add_argument_group(
        title="Download",
    )
    add_asset_dir_argument(download_group)
    add_platform_argument(download_group)
    download_group.add_argument(
        "--overwrite",
        action="store_true",
        help="Allows downloading into an existing app directory and potentially overwriting existing files.",
    )
    download_group.add_argument(
        "--version-state",
        choices=list(x.name for x in appstore.VersionState),
        metavar="VERSION_STATE",
        help="Specify the required appstore version state.  The first matching version will be downloaded. "
        + "By default, the first version listed by the app store is used.",
    )
    add_authentication_group(download_parser)
    add_app_id_group(download_parser)

    # Action: publish
    publish_parser = add_subparser(
        action_subparsers,
        "publish",
        help="Publish all assets for an app.",
    )
    publish_group = publish_parser.add_argument_group(
        title="Publish",
    )
    add_asset_dir_argument(publish_group)
    add_platform_argument(publish_group)
    publish_group.add_argument(
        "--version-string",
        help="Update the version string in the app store version.",
    )
    publish_group.add_argument(
        "--created-version-string",
        default="NEW.APP.VERSION",
        help=f"If {clr_keyword('--version-string')} isn't specified, "
        + "use this as the version string when creating a new app store version.",
    )
    publish_group.add_argument(
        "--no-create-version",
        action="store_true",
        help="Prevent a new app version from being created when no versions are in an editable state.",
    )
    publish_group.add_argument(
        "--no-create-locale",
        action="store_true",
        help="Prevent new locales from being created.",
    )
    publish_group.add_argument(
        "--no-delete-locale",
        action="store_true",
        help="Prevent locales from being deleted.",
    )
    add_authentication_group(publish_parser)
    add_app_id_group(publish_parser)

    # Parse
    parsed_args = global_parser.parse_known_args()

    # tuple { matched_args, remaining_args }
    args = parsed_args[0]

    # Handle loading the auth key from file
    if "key" in args and "key_file" in args and args.key == None:
        args.key = args.key_file.read()
        args.key_file.close()

    # Set LogLevel
    logging.getLogger().setLevel(args.log_level)

    # Run
    try:
        if args.action == "categories":
            console_actions.list_categories(args)
        if args.action == "apps":
            console_actions.list_apps(args)
        elif args.action == "infos":
            console_actions.list_infos(args)
        elif args.action == "versions":
            console_actions.list_versions(args)
        elif args.action == "screenshots":
            console_actions.list_screenshots(args)
        elif args.action == "previews":
            console_actions.list_previews(args)
        elif args.action == "download":
            console_actions.download(args)
        elif args.action == "publish":
            console_actions.publish(args)
    except requests.exceptions.SSLError as error:
        sys.exit(error)
    except requests.exceptions.ConnectionError as error:
        sys.exit(error)
    except requests.exceptions.HTTPError as error:
        sys.exit(error)
    except appstore.ResourceNotFoundException as error:
        sys.exit(error)
    except FileExistsError as error:
        sys.exit(error)
    except KeyboardInterrupt as error:
        sys.exit(error)



import requests
import json
import gzip
import colorama
import logging
from enum import Enum, auto
from appstore_tools.print_util import json_term, clr
from .types import (
    Platform,
    PlatformList,
    ScreenshotDisplayType,
    PreviewType,
    VersionState,
    VersionStateList,
    InfoAttributes,
    InfoLocalizationAttributes,
    VersionAttributes,
    VersionLocalizationAttributes,
)
from .util import enum_name, enum_names, editable_version_states
from .fetch import fetch
from .exceptions import ResourceNotFoundException
from .fetch import fetch, FetchMethod

# TODO: remove pylint "disable" directives when pylint supports python 3.9 completely
from typing import TypedDict, Optional, Union, Literal, Sequence


def get_categories(
    access_token: str,
    platforms: PlatformList = list(Platform),
):
    """Get this list of possible categories/subcategories on the app store."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appCategories?filter[platforms]={','.join(enum_names(platforms))}&exists[parent]=false&include=subcategories",
        access_token=access_token,
    )["data"]


def get_apps(
    access_token: str,
):
    """Get all apps under the users app store account."""
    return fetch(method=FetchMethod.GET, path=f"/apps", access_token=access_token)[
        "data"
    ]


def get_app(
    app_id: str,
    access_token: str,
):
    """Get app by id."""
    return fetch(
        method=FetchMethod.GET, path=f"/apps/{app_id}", access_token=access_token
    )["data"]


def get_app_id(
    bundle_id: str,
    access_token: str,
) -> int:
    """Get the app id for the specified bundle id."""
    apps = get_apps(access_token)
    try:
        app_id = next(
            app["id"] for app in apps if app["attributes"]["bundleId"] == bundle_id
        )
        return int(app_id)
    except StopIteration:
        raise ResourceNotFoundException(f'No app matching bundle-id "{bundle_id}"')


def get_bundle_id(
    app_id: str,
    access_token: str,
) -> int:
    """Get the bundle id for the specified app id."""
    app = get_app(app_id=app_id, access_token=access_token)
    return app["attributes"]["bundleId"]


def get_infos(
    app_id: str,
    access_token: str,
    states: VersionStateList = list(VersionState),
):
    """Get the list of app infos, optionally filtering by appstore state."""
    versions = fetch(
        method=FetchMethod.GET,
        path=f"/apps/{app_id}/appInfos",
        access_token=access_token,
    )["data"]

    return [
        v for v in versions if v["attributes"]["appStoreState"] in enum_names(states)
    ]


def update_info(
    info_id: str,
    info_attributes: InfoAttributes,
    access_token: str,
):
    """Update the non-localized AppInfo data."""
    relationships = {}
    for k, v in info_attributes.items():
        relationships[k] = {"data": {"id": v, "type": "appCategories"}}

    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appInfos/{info_id}",
        access_token=access_token,
        data={
            "data": {
                "id": info_id,
                "relationships": relationships,
                "type": "appInfos",
            }
        },
    )["data"]


def get_info_localizations(
    info_id: str,
    access_token: str,
):
    """Get the list of app info localizations."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appInfos/{info_id}/appInfoLocalizations",
        access_token=access_token,
    )["data"]


def create_info_localization(
    info_id: str,
    locale: str,
    info_localization_attributes: InfoLocalizationAttributes,
    access_token: str,
):
    """Creates a new app store info localization."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appInfoLocalizations",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"locale": locale, **info_localization_attributes},
                "relationships": {
                    "appInfo": {"data": {"id": info_id, "type": "appInfos"}}
                },
                "type": "appInfoLocalizations",
            }
        },
    )["data"]


def update_info_localization(
    info_localization_id: str,
    info_localization_attributes: InfoLocalizationAttributes,
    access_token: str,
):
    """Updates the meta data for the specified App Info Localization.
    Some data fields require the App Version to be in an editable state."""
    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appInfoLocalizations/{info_localization_id}",
        access_token=access_token,
        data={
            "data": {
                "id": info_localization_id,
                "attributes": info_localization_attributes,
                "type": "appInfoLocalizations",
            }
        },
    )["data"]


def delete_info_localization(
    info_localization_id: str,
    access_token: str,
):
    """Deletes the specified App Info Localization."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appInfoLocalizations/{info_localization_id}",
        access_token=access_token,
    )


def create_version(
    app_id: str,
    platform: Union[Platform, str],  # pylint: disable=unsubscriptable-object
    version_string: str,
    access_token: str,
):
    """Creates a new app version."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appStoreVersions/",
        access_token=access_token,
        data={
            "data": {
                "attributes": {
                    "platform": enum_name(platform),
                    "versionString": version_string,
                },
                "relationships": {"app": {"data": {"id": app_id, "type": "apps"}}},
                "type": "appStoreVersions",
            }
        },
    )["data"]


def update_version(
    version_id: str,
    version_attributes: VersionAttributes,
    access_token: str,
):
    """Update an app version."""
    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appStoreVersions/{version_id}",
        access_token=access_token,
        data={
            "data": {
                "id": version_id,
                "attributes": version_attributes,
                "type": "appStoreVersions",
            }
        },
    )["data"]


def get_versions(
    app_id: str,
    access_token: str,
    platforms: PlatformList = list(Platform),
    states: VersionStateList = list(VersionState),
):
    """Get the list of app versions, optionally filtering by platform and/or state."""
    versions = fetch(
        method=FetchMethod.GET,
        path=f"/apps/{app_id}/appStoreVersions",
        access_token=access_token,
    )["data"]

    return [
        v
        for v in versions
        if v["attributes"]["platform"] in enum_names(platforms)
        and v["attributes"]["appStoreState"] in enum_names(states)
    ]


def get_versions_editable(
    app_id: str,
    access_token: str,
    platforms: PlatformList = list(Platform),
):
    return get_versions(
        app_id=app_id,
        access_token=access_token,
        platforms=platforms,
        states=[s.name for s in editable_version_states],
    )


def get_version_live(
    app_id: str,
    access_token: str,
    platforms: PlatformList = list(Platform),
):
    live_state = VersionState.READY_FOR_SALE.name
    versions = get_versions(
        app_id=app_id,
        access_token=access_token,
        platforms=platforms,
        states=[live_state],
    )

    if len(versions) == 0:
        raise ResourceNotFoundException(f'No app version matching state "{live_state}"')
    else:
        return versions[0]


def get_version_localizations(
    version_id: str,
    access_token: str,
):
    return fetch(
        method=FetchMethod.GET,
        path=f"/appStoreVersions/{version_id}/appStoreVersionLocalizations",
        access_token=access_token,
    )["data"]


def create_version_localization(
    version_id: str,
    locale: str,
    localization_attributes: VersionLocalizationAttributes,
    access_token: str,
):
    """Creates a new app store version localization."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appStoreVersionLocalizations",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"locale": locale, **localization_attributes},
                "relationships": {
                    "appStoreVersion": {
                        "data": {"id": version_id, "type": "appStoreVersions"}
                    }
                },
                "type": "appStoreVersionLocalizations",
            }
        },
    )["data"]


def update_version_localization(
    localization_id: str,
    localization_attributes: VersionLocalizationAttributes,
    access_token: str,
):
    """Updates the meta data for the specified App Version Localization.
    Some data fields require the App Version to be in an editable state."""
    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appStoreVersionLocalizations/{localization_id}",
        access_token=access_token,
        data={
            "data": {
                "id": localization_id,
                "attributes": localization_attributes,
                "type": "appStoreVersionLocalizations",
            }
        },
    )["data"]


def delete_version_localization(
    localization_id: str,
    access_token: str,
):
    """Deletes the specified App Version Localization."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appStoreVersionLocalizations/{localization_id}",
        access_token=access_token,
    )


def get_screenshot_sets(
    localization_id: str,
    access_token: str,
):
    """Get the screenshot sets from the specified App Version Localization."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appStoreVersionLocalizations/{localization_id}/appScreenshotSets",
        access_token=access_token,
    )["data"]


def create_screenshot_set(
    localization_id: str,
    display_type: Union[
        ScreenshotDisplayType, str
    ],  # pylint: disable=unsubscriptable-object
    access_token: str,
):
    """Create a new screenshot set in the specified App Version Localization."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appScreenshotSets",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"screenshotDisplayType": enum_name(display_type)},
                "relationships": {
                    "appStoreVersionLocalization": {
                        "data": {
                            "id": localization_id,
                            "type": "appStoreVersionLocalizations",
                        }
                    }
                },
                "type": "appScreenshotSets",
            }
        },
    )["data"]


def delete_screenshot_set(
    screenshot_set_id: str,
    access_token: str,
):
    """Delete a screenshot set from the App Version Localization."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appScreenshotSets/{screenshot_set_id}",
        access_token=access_token,
    )


def update_screenshot_order(
    screenshot_set_id: str,
    screenshot_ids: Sequence[str],
    access_token: str,
):
    """Update the order of the screenshots in a screenshot set."""
    fetch(
        method=FetchMethod.PATCH,
        path=f"/appScreenshotSets/{screenshot_set_id}/relationships/appScreenshots",
        access_token=access_token,
        data={
            "data": [
                {"id": ss_id, "type": "appScreenshots"} for ss_id in screenshot_ids
            ]
        },
    )


def get_screenshots(
    screenshot_set_id: str,
    access_token: str,
):
    """Get the screenshots in a screenshot set."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appScreenshotSets/{screenshot_set_id}/appScreenshots",
        access_token=access_token,
    )["data"]


def get_screenshot(
    screenshot_id: str,
    access_token: str,
):
    """Get the screenshot info."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appScreenshots/{screenshot_id}",
        access_token=access_token,
    )["data"]


def create_screenshot(
    screenshot_set_id: str,
    file_name: str,
    file_size: int,
    access_token: str,
):
    """Create a screenshot asset reservation in the specified screenshot set.
    Use the upload operations in the response to upload the file parts."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appScreenshots",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"fileName": file_name, "fileSize": file_size},
                "relationships": {
                    "appScreenshotSet": {
                        "data": {
                            "id": screenshot_set_id,
                            "type": "appScreenshotSets",
                        }
                    }
                },
                "type": "appScreenshots",
            }
        },
    )["data"]


def update_screenshot(
    screenshot_id: str,
    uploaded: bool,
    sourceFileChecksum: str,
    access_token: str,
):
    """Update the screenshot to commit it after a successful upload."""
    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appScreenshots/{screenshot_id}",
        access_token=access_token,
        data={
            "data": {
                "id": screenshot_id,
                "attributes": {
                    "uploaded": uploaded,
                    "sourceFileChecksum": sourceFileChecksum,
                },
                "type": "appScreenshots",
            }
        },
    )["data"]


def delete_screenshot(
    screenshot_id: str,
    access_token: str,
):
    """Delete a screenshot from its screenshot set."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appScreenshots/{screenshot_id}",
        access_token=access_token,
    )


def get_preview_sets(
    localization_id: str,
    access_token: str,
):
    """Get the preview sets in the specified App Version Localization."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appStoreVersionLocalizations/{localization_id}/appPreviewSets",
        access_token=access_token,
    )["data"]


def create_preview_set(
    localization_id: str,
    preview_type: Union[PreviewType, str],  # pylint: disable=unsubscriptable-object
    access_token: str,
):
    """Create a new preview set in the specified App Version Localization."""
    return fetch(
        method=FetchMethod.POST,
        path=f"/appPreviewSets",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"previewType": enum_name(preview_type)},
                "relationships": {
                    "appStoreVersionLocalization": {
                        "data": {
                            "id": localization_id,
                            "type": "appStoreVersionLocalizations",
                        }
                    }
                },
                "type": "appPreviewSets",
            }
        },
    )["data"]


def delete_preview_set(
    preview_set_id: str,
    access_token: str,
):
    """Delete a preview set from the App Version Localization."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appPreviewSets/{preview_set_id}",
        access_token=access_token,
    )


def update_preview_order(
    preview_set_id: str,
    preview_ids: Sequence[str],
    access_token: str,
):
    """Update the order of the previews in a preview set."""
    fetch(
        method=FetchMethod.PATCH,
        path=f"/appPreviewSets/{preview_set_id}/relationships/appPreviews",
        access_token=access_token,
        data={"data": [{"id": p_id, "type": "appPreviews"} for p_id in preview_ids]},
    )


def get_previews(
    preview_set_id: str,
    access_token: str,
):
    """Get the previews in a preview set."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appPreviewSets/{preview_set_id}/appPreviews",
        access_token=access_token,
    )["data"]


def get_preview(
    preview_id: str,
    access_token: str,
):
    """Get the preview info."""
    return fetch(
        method=FetchMethod.GET,
        path=f"/appPreviews/{preview_id}",
        access_token=access_token,
    )["data"]


def create_preview(
    preview_set_id: str,
    file_name: str,
    file_size: int,
    access_token: str,
):
    """Create a preview asset reservation in the specified preview set.
    Use the upload operations in the response to upload the file parts."""

    # TODO: add support for previewFrameTimeCode
    return fetch(
        method=FetchMethod.POST,
        path=f"/appPreviews",
        access_token=access_token,
        data={
            "data": {
                "attributes": {"fileName": file_name, "fileSize": file_size},
                "relationships": {
                    "appPreviewSet": {
                        "data": {
                            "id": preview_set_id,
                            "type": "appPreviewSets",
                        }
                    }
                },
                "type": "appPreviews",
            }
        },
    )["data"]


def update_preview(
    preview_id: str,
    uploaded: bool,
    sourceFileChecksum: str,
    access_token: str,
):
    """Update the preview to commit it after a successful upload."""

    # TODO: add support for previewFrameTimeCode
    return fetch(
        method=FetchMethod.PATCH,
        path=f"/appPreviews/{preview_id}",
        access_token=access_token,
        data={
            "data": {
                "id": preview_id,
                "attributes": {
                    "uploaded": uploaded,
                    "sourceFileChecksum": sourceFileChecksum,
                },
                "type": "appPreviews",
            }
        },
    )["data"]


def delete_preview(
    preview_id: str,
    access_token: str,
):
    """Delete a preview from its preview set."""
    fetch(
        method=FetchMethod.DELETE,
        path=f"/appPreviews/{preview_id}",
        access_token=access_token,
    )



from .list import (
    list_apps,
    list_categories,
    list_infos,
    list_previews,
    list_screenshots,
    list_versions,
    Verbosity,
)
from .download import download
from .publish import publish


from enum import Enum, auto

# TODO: remove pylint "disable" directives when pylint supports python 3.9 completely
from typing import TypedDict, Optional, Union, Literal, Sequence


class Platform(Enum):
    IOS = auto()
    MAC_OS = auto()
    TV_OS = auto()


class ReleaseType(Enum):
    MANUAL = auto()
    AFTER_APPROVAL = auto()
    SCHEDULED = auto()


class VersionState(Enum):
    DEVELOPER_REMOVED_FROM_SALE = auto()
    DEVELOPER_REJECTED = auto()
    IN_REVIEW = auto()
    INVALID_BINARY = auto()
    METADATA_REJECTED = auto()
    PENDING_APPLE_RELEASE = auto()
    PENDING_CONTRACT = auto()
    PENDING_DEVELOPER_RELEASE = auto()
    PREPARE_FOR_SUBMISSION = auto()
    PREORDER_READY_FOR_SALE = auto()
    PROCESSING_FOR_APP_STORE = auto()
    READY_FOR_SALE = auto()
    REJECTED = auto()
    REMOVED_FROM_SALE = auto()
    WAITING_FOR_EXPORT_COMPLIANCE = auto()
    WAITING_FOR_REVIEW = auto()
    REPLACED_WITH_NEW_VERSION = auto()


class ScreenshotDisplayType(Enum):
    APP_IPHONE_65 = auto()
    APP_IPHONE_58 = auto()
    APP_IPHONE_55 = auto()
    APP_IPHONE_47 = auto()
    APP_IPHONE_40 = auto()
    APP_IPHONE_35 = auto()
    APP_IPAD_PRO_3GEN_129 = auto()
    APP_IPAD_PRO_3GEN_11 = auto()
    APP_IPAD_PRO_129 = auto()
    APP_IPAD_105 = auto()
    APP_IPAD_97 = auto()
    APP_DESKTOP = auto()
    APP_WATCH_SERIES_4 = auto()
    APP_WATCH_SERIES_3 = auto()
    APP_APPLE_TV = auto()
    IMESSAGE_APP_IPHONE_65 = auto()
    IMESSAGE_APP_IPHONE_58 = auto()
    IMESSAGE_APP_IPHONE_55 = auto()
    IMESSAGE_APP_IPHONE_47 = auto()
    IMESSAGE_APP_IPHONE_40 = auto()
    IMESSAGE_APP_IPAD_PRO_3GEN_129 = auto()
    IMESSAGE_APP_IPAD_PRO_3GEN_11 = auto()
    IMESSAGE_APP_IPAD_PRO_129 = auto()
    IMESSAGE_APP_IPAD_105 = auto()
    IMESSAGE_APP_IPAD_97 = auto()


class PreviewType(Enum):
    IPHONE_65 = auto()
    IPHONE_58 = auto()
    IPHONE_55 = auto()
    IPHONE_47 = auto()
    IPHONE_40 = auto()
    IPHONE_35 = auto()
    IPAD_PRO_3GEN_129 = auto()
    IPAD_PRO_3GEN_11 = auto()
    IPAD_PRO_129 = auto()
    IPAD_105 = auto()
    IPAD_97 = auto()
    DESKTOP = auto()
    WATCH_SERIES_4 = auto()
    WATCH_SERIES_3 = auto()
    APPLE_TV = auto()


class MediaAssetState(Enum):
    AWAITING_UPLOAD = auto()
    UPLOAD_COMPLETE = auto()
    COMPLETE = auto()
    FAILED = auto()


EnumList = Sequence[Union[Enum, str]]  # pylint: disable=unsubscriptable-object
PlatformList = Sequence[Union[Platform, str]]  # pylint: disable=unsubscriptable-object
VersionStateList = Sequence[
    Union[VersionState, str]  # pylint: disable=unsubscriptable-object
]


class InfoAttributes(TypedDict, total=False):  # pylint: disable=inherit-non-class
    primaryCategory: Optional[str]  # pylint: disable=unsubscriptable-object
    primarySubcategoryOne: Optional[str]  # pylint: disable=unsubscriptable-object
    primarySubcategoryTwo: Optional[str]  # pylint: disable=unsubscriptable-object
    secondaryCategory: Optional[str]  # pylint: disable=unsubscriptable-object
    secondarySubcategoryOne: Optional[str]  # pylint: disable=unsubscriptable-object
    secondarySubcategoryTwo: Optional[str]  # pylint: disable=unsubscriptable-object


class InfoLocalizationAttributes(
    TypedDict, total=False
):  # pylint: disable=inherit-non-class
    name: Optional[str]  # pylint: disable=unsubscriptable-object
    privacyPolicyText: Optional[str]  # pylint: disable=unsubscriptable-object
    privacyPolicyUrl: Optional[str]  # pylint: disable=unsubscriptable-object
    subtitle: Optional[str]  # pylint: disable=unsubscriptable-object


class VersionAttributes(TypedDict, total=False):  # pylint: disable=inherit-non-class
    copyright: Optional[str]  # pylint: disable=unsubscriptable-object
    earliestReleaseDate: Optional[str]  # pylint: disable=unsubscriptable-object
    releaseType: Optional[str]  # pylint: disable=unsubscriptable-object
    usesIdfa: Optional[bool]  # pylint: disable=unsubscriptable-object
    versionString: Optional[str]  # pylint: disable=unsubscriptable-object
    downloadable: Optional[bool]  # pylint: disable=unsubscriptable-object


class VersionLocalizationAttributes(
    TypedDict, total=False
):  # pylint: disable=inherit-non-class
    description: Optional[str]  # pylint: disable=unsubscriptable-object
    keywords: Optional[str]  # pylint: disable=unsubscriptable-object
    marketingUrl: Optional[str]  # pylint: disable=unsubscriptable-object
    promotionalText: Optional[str]  # pylint: disable=unsubscriptable-object
    supportUrl: Optional[str]  # pylint: disable=unsubscriptable-object
    whatsNew: Optional[str]  # pylint: disable=unsubscriptable-object



import os
import hashlib
import colorama
import requests
from typing import Union, Sequence
from appstore_tools import appstore
from appstore_tools.print_util import print_clr, clr, json_term
from .util import (
    read_txt_file,
    print_locale_status,
    print_media_set_status,
    print_media_status,
)


def media_checksum_ok(media, media_asset_dir: str) -> bool:
    """Checks if the appstore checksum matches the asset checksum."""
    file_name = media["attributes"]["fileName"]
    file_path = os.path.join(media_asset_dir, file_name)
    appstore_checksum = media["attributes"]["sourceFileChecksum"]

    if appstore_checksum is None:
        print_media_status(
            file_name,
            colorama.Fore.CYAN,
            "checksum missing (in processing)",
        )
        return False

    if not os.path.isfile(file_path):
        print_media_status(
            file_name,
            colorama.Fore.RED,
            "no source file",
        )
        return False

    with open(file_path, "rb") as file:
        asset_checksum = hashlib.md5(file.read()).hexdigest()
        if asset_checksum == appstore_checksum:
            print_media_status(
                file_name,
                colorama.Fore.CYAN + colorama.Style.DIM,
                clr(
                    f"checksum matched: ",
                    f"{colorama.Style.DIM}{asset_checksum}",
                ),
            )
        else:
            print_media_status(
                file_name,
                colorama.Fore.CYAN,
                clr(
                    f"checksum changed: ",
                    f"{colorama.Style.DIM}{appstore_checksum} -> {asset_checksum}",
                ),
            )
        return asset_checksum == appstore_checksum


def upload_media(media, media_asset_path: str) -> str:
    """Upload media asset (screenshot or preview) to the appstore.

    Returns:
        str: checksum
    """
    file_hash = hashlib.md5()
    upload_operations = media["attributes"]["uploadOperations"]

    for op in upload_operations:
        method: str = op["method"]
        url: str = op["url"]
        headers: dict = {}
        for h in op["requestHeaders"]:
            headers[h["name"]] = h["value"]
        length: int = op["length"]
        offset: int = op["offset"]

        with open(media_asset_path, "rb") as file:
            file.seek(offset)
            file_chunk = file.read(length)

        file_hash.update(file_chunk)
        print_media_status(
            media_asset_path,
            colorama.Fore.CYAN,
            f"uploading chunk (offset: {offset}, length: {length})",
        )
        requests.request(method=method, url=url, headers=headers, data=file_chunk)
    return file_hash.hexdigest()


def get_media_file_names(media: Sequence[dict]):
    return [x["attributes"]["fileName"] for x in media]


def get_asset_file_names(asset_dir: str):
    return [
        x for x in os.listdir(asset_dir) if os.path.isfile(os.path.join(asset_dir, x))
    ]


def get_new_file_paths(media: Sequence[dict], asset_dir: str):
    media_file_names = get_media_file_names(media)
    asset_file_names = get_asset_file_names(asset_dir)
    return [
        os.path.join(asset_dir, x)
        for x in asset_file_names
        if x not in media_file_names
    ]


def publish_screenshot(
    access_token: str,
    screenshot_path: str,
    screenshot_set_id: str,
):
    if not os.path.isfile(screenshot_path):
        raise FileNotFoundError(f"Screenshot path does not exist: {screenshot_path}")

    _, file_name = os.path.split(screenshot_path)
    file_stat = os.stat(screenshot_path)

    # Create
    print_media_status(
        file_name,
        colorama.Fore.CYAN,
        "reserving asset",
    )
    screenshot = appstore.create_screenshot(
        screenshot_set_id=screenshot_set_id,
        file_name=file_name,
        file_size=file_stat.st_size,
        access_token=access_token,
    )
    # Upload
    checksum = upload_media(media=screenshot, media_asset_path=screenshot_path)

    # Commit
    print_media_status(
        file_name,
        colorama.Fore.CYAN,
        "commiting upload",
    )
    screenshot = appstore.update_screenshot(
        screenshot_id=screenshot["id"],
        uploaded=True,
        sourceFileChecksum=checksum,
        access_token=access_token,
    )


def publish_screenshots(
    access_token: str,
    screenshot_set_dir: str,
    screenshot_set_id: str,
    display_type: str,
):
    print_media_set_status(display_type, colorama.Fore.CYAN, "checking for changes")

    # Delete outdated screenshots
    screenshots = appstore.get_screenshots(
        screenshot_set_id=screenshot_set_id, access_token=access_token
    )
    for screenshot in screenshots:
        if not media_checksum_ok(media=screenshot, media_asset_dir=screenshot_set_dir):
            appstore.delete_screenshot(
                screenshot_id=screenshot["id"], access_token=access_token
            )

    # Create new screenshots
    screenshots = appstore.get_screenshots(
        screenshot_set_id=screenshot_set_id, access_token=access_token
    )

    # Publish
    new_file_paths = get_new_file_paths(screenshots, screenshot_set_dir)
    for file_path in new_file_paths:
        publish_screenshot(
            access_token=access_token,
            screenshot_path=file_path,
            screenshot_set_id=screenshot_set_id,
        )

    # Reorder the screenshots
    print_media_set_status(display_type, colorama.Fore.CYAN, "sorting screenshots")
    screenshots = appstore.get_screenshots(
        screenshot_set_id=screenshot_set_id, access_token=access_token
    )
    screenshots.sort(key=lambda x: x["attributes"]["fileName"])
    screenshot_ids = [x["id"] for x in screenshots]
    appstore.update_screenshot_order(
        screenshot_set_id=screenshot_set_id,
        screenshot_ids=screenshot_ids,
        access_token=access_token,
    )


def publish_screenshot_sets(
    access_token: str,
    localization_dir: str,
    localization_id: str,
):
    """Publish the screenshot sets from assets on disk."""
    screenshots_dir = os.path.join(localization_dir, "screenshots")
    if not os.path.isdir(screenshots_dir):
        print_clr(
            f"    No screenshots: directory {colorama.Fore.CYAN}{screenshots_dir}{colorama.Fore.RESET} not found.",
        )
        return

    screenshot_sets = appstore.get_screenshot_sets(
        localization_id=localization_id, access_token=access_token
    )

    asset_display_types = [
        x
        for x in os.listdir(screenshots_dir)
        if os.path.isdir(os.path.join(screenshots_dir, x))
    ]

    # Create new display types
    loc_display_types = [
        x["attributes"]["screenshotDisplayType"] for x in screenshot_sets
    ]
    new_display_types = [x for x in asset_display_types if x not in loc_display_types]
    for display_type in new_display_types:
        print_media_set_status(
            display_type, colorama.Fore.YELLOW, "creating display type"
        )
        screenshot_set = appstore.create_screenshot_set(
            localization_id=localization_id,
            display_type=display_type,
            access_token=access_token,
        )
        screenshot_sets.append(screenshot_set)

    for screenshot_set in screenshot_sets:
        screenshot_set_id = screenshot_set["id"]
        display_type = screenshot_set["attributes"]["screenshotDisplayType"]
        screenshot_set_dir = os.path.join(screenshots_dir, display_type)

        # Delete removed display types
        if not os.path.isdir(screenshot_set_dir):
            print_media_set_status(
                display_type, colorama.Fore.RED, "deleting display type"
            )
            appstore.delete_screenshot_set(
                screenshot_set_id=screenshot_set_id, access_token=access_token
            )
            continue

        # Publish
        publish_screenshots(
            access_token=access_token,
            screenshot_set_dir=screenshot_set_dir,
            screenshot_set_id=screenshot_set_id,
            display_type=display_type,
        )


def publish_preview(
    access_token: str,
    preview_path: str,
    preview_set_id: str,
):
    if not os.path.isfile(preview_path):
        raise FileNotFoundError(f"Preview path does not exist: {preview_path}")

    _, file_name = os.path.split(preview_path)
    file_stat = os.stat(preview_path)

    # Create
    print_media_status(
        file_name,
        colorama.Fore.CYAN,
        "reserving asset",
    )
    preview = appstore.create_preview(
        preview_set_id=preview_set_id,
        file_name=file_name,
        file_size=file_stat.st_size,
        access_token=access_token,
    )
    # Upload
    checksum = upload_media(media=preview, media_asset_path=preview_path)

    # Commit
    print_media_status(
        file_name,
        colorama.Fore.CYAN,
        "commiting upload",
    )
    preview = appstore.update_preview(
        preview_id=preview["id"],
        uploaded=True,
        sourceFileChecksum=checksum,
        access_token=access_token,
    )


def publish_previews(
    access_token: str,
    preview_set_dir: str,
    preview_set_id: str,
    display_type: str,
):
    print_media_set_status(display_type, colorama.Fore.CYAN, "checking for changes")

    # Delete outdated previews
    previews = appstore.get_previews(
        preview_set_id=preview_set_id, access_token=access_token
    )
    for preview in previews:
        if not media_checksum_ok(media=preview, media_asset_dir=preview_set_dir):
            appstore.delete_preview(preview_id=preview["id"], access_token=access_token)

    # Create new previews
    previews = appstore.get_previews(
        preview_set_id=preview_set_id, access_token=access_token
    )
    new_file_paths = get_new_file_paths(previews, preview_set_dir)

    # Publish
    for file_path in new_file_paths:
        publish_preview(
            access_token=access_token,
            preview_path=file_path,
            preview_set_id=preview_set_id,
        )

    # Reorder the previews
    print_media_set_status(display_type, colorama.Fore.CYAN, "sorting previews")
    previews = appstore.get_previews(
        preview_set_id=preview_set_id, access_token=access_token
    )
    previews.sort(key=lambda x: x["attributes"]["fileName"])
    preview_ids = [x["id"] for x in previews]
    appstore.update_preview_order(
        preview_set_id=preview_set_id,
        preview_ids=preview_ids,
        access_token=access_token,
    )


def publish_preview_sets(
    access_token: str,
    localization_dir: str,
    localization_id: str,
):
    """Publish the previews sets from assets on disk."""
    previews_dir = os.path.join(localization_dir, "previews")
    if not os.path.isdir(previews_dir):
        print_clr(
            f"    No previews: directory {colorama.Fore.CYAN}{previews_dir}{colorama.Fore.RESET} not found.",
        )
        return

    preview_sets = appstore.get_preview_sets(
        localization_id=localization_id, access_token=access_token
    )

    asset_display_types = [
        x
        for x in os.listdir(previews_dir)
        if os.path.isdir(os.path.join(previews_dir, x))
    ]

    # Create new display types
    loc_preview_types = [x["attributes"]["previewType"] for x in preview_sets]
    new_preview_types = [x for x in asset_display_types if x not in loc_preview_types]
    for preview_type in new_preview_types:
        print_media_set_status(
            preview_type, colorama.Fore.YELLOW, "creating preview type"
        )
        preview_set = appstore.create_preview_set(
            localization_id=localization_id,
            preview_type=preview_type,
            access_token=access_token,
        )
        preview_sets.append(preview_set)

    for preview_set in preview_sets:
        preview_set_id = preview_set["id"]
        preview_type = preview_set["attributes"]["previewType"]
        preview_set_dir = os.path.join(previews_dir, preview_type)

        # Delete removed display types
        if not os.path.isdir(preview_set_dir):
            print_media_set_status(
                preview_type, colorama.Fore.RED, "deleting preview type"
            )
            appstore.delete_preview_set(
                preview_set_id=preview_set_id, access_token=access_token
            )
            continue

        # Publish
        publish_previews(
            access_token=access_token,
            preview_set_dir=preview_set_dir,
            preview_set_id=preview_set_id,
            display_type=preview_type,
        )


def publish_version_localizations(
    access_token: str,
    app_dir: str,
    version_id: str,
    allow_create_locale: bool = True,
    allow_delete_locale: bool = True,
):
    localizations = appstore.get_version_localizations(
        version_id=version_id, access_token=access_token
    )

    asset_locales = [
        x for x in os.listdir(app_dir) if os.path.isdir(os.path.join(app_dir, x))
    ]

    # create new localizations
    version_locales = [loc["attributes"]["locale"] for loc in localizations]
    new_locales = [x for x in asset_locales if x not in version_locales]
    if allow_create_locale:
        for locale in new_locales:
            print_locale_status(locale, colorama.Fore.YELLOW, "creating locale")
            loc = appstore.create_version_localization(
                version_id=version_id,
                locale=locale,
                localization_attributes={},
                access_token=access_token,
            )
            localizations.append(loc)
    else:
        for locale in new_locales:
            print_locale_status(
                locale, colorama.Fore.LIGHTBLACK_EX, "locale creation not allowed"
            )

    # publish localizations
    for loc in localizations:
        loc_id = loc["id"]
        loc_attr = loc["attributes"]
        locale = loc_attr["locale"]
        loc_dir = os.path.join(app_dir, locale)

        # Delete removed locales
        if not os.path.isdir(loc_dir):
            if allow_delete_locale:
                print_locale_status(locale, colorama.Fore.RED, "deleting locale")
                appstore.delete_version_localization(
                    localization_id=loc_id, access_token=access_token
                )
            else:
                print_locale_status(
                    locale, colorama.Fore.LIGHTBLACK_EX, "locale deletion not allowed"
                )
            continue

        # Normalize all attribute values to strings
        for key in appstore.VersionLocalizationAttributes.__annotations__.keys():
            if loc_attr[key] is None:
                loc_attr[key] = ""

        # Load local data from disk
        asset_loc_data: appstore.VersionLocalizationAttributes = {}
        for key in appstore.VersionLocalizationAttributes.__annotations__.keys():
            path = os.path.join(loc_dir, key + ".txt")
            content = read_txt_file(path)
            if content is not None:
                asset_loc_data[key] = content  # type: ignore

        # Only need to update if there are differences
        loc_diff_keys = [
            key
            for key, value in asset_loc_data.items()
            if value is not None and value != loc_attr[key]
        ]
        if len(loc_diff_keys) > 0:
            print_locale_status(
                locale,
                colorama.Fore.CYAN,
                f"updating locale {colorama.Fore.CYAN}{colorama.Style.DIM}{loc_diff_keys}",
            )
            appstore.update_version_localization(
                localization_id=loc_id,
                localization_attributes=asset_loc_data,
                access_token=access_token,
            )
        else:
            print_locale_status(
                locale, colorama.Fore.CYAN, "no changes in version settings"
            )

        # Screenshots
        publish_screenshot_sets(
            access_token=access_token,
            localization_dir=loc_dir,
            localization_id=loc_id,
        )

        # Previews
        publish_preview_sets(
            access_token=access_token,
            localization_dir=loc_dir,
            localization_id=loc_id,
        )


def publish_version(
    access_token: str,
    app_dir: str,
    app_id: str,
    bundle_id: str,
    platform: Union[appstore.Platform, str],  # pylint: disable=unsubscriptable-object
    version_string: str,
    update_version_string: bool,
    allow_create_version: bool = True,
    allow_create_locale: bool = True,
    allow_delete_locale: bool = True,
):
    # Get Versions
    versions = appstore.get_versions(
        app_id=app_id,
        access_token=access_token,
        platforms=[platform],
        states=appstore.editable_version_states,
    )
    print_clr(
        f"Found {colorama.Fore.CYAN}{len(versions)}{colorama.Fore.RESET} editable app versions ",
        f"for {colorama.Fore.CYAN}{platform}{colorama.Fore.RESET}.",
    )

    if len(versions) == 0 and allow_create_version:
        print(
            f"Creating new version: {colorama.Fore.BLUE}{version_string}{colorama.Fore.RESET}"
        )
        created_version = appstore.create_version(
            app_id=app_id,
            platform=platform,
            version_string=version_string,
            access_token=access_token,
        )
        versions.append(created_version)
    elif update_version_string:
        for v in versions:
            version_id = v["id"]
            version_state = v["attributes"]["appStoreState"]

            version_attributes: appstore.VersionAttributes = {
                "versionString": version_string,
            }
            print_clr(
                f"{colorama.Fore.GREEN}Version ",
                f"{colorama.Fore.BLUE}{version_state} ",
                f": updating version ",
                f"{colorama.Fore.CYAN}{version_attributes}",
            )

            appstore.update_version(
                version_id=version_id,
                version_attributes=version_attributes,
                access_token=access_token,
            )

    for v in versions:
        version_id = v["id"]
        version_state = v["attributes"]["appStoreState"]

        print_clr(
            f"{colorama.Fore.GREEN}Version ",
            f"{colorama.Fore.BLUE}{version_id} ",
            f"{colorama.Fore.CYAN}{version_state} ",
        )
        publish_version_localizations(
            access_token=access_token,
            app_dir=app_dir,
            version_id=version_id,
            allow_create_locale=allow_create_locale,
            allow_delete_locale=allow_delete_locale,
        )


def publish_info(
    access_token: str,
    app_dir: str,
    app_id: str,
    bundle_id: str,
    platform: Union[appstore.Platform, str],  # pylint: disable=unsubscriptable-object
):
    # Get Infos
    infos = appstore.get_infos(
        app_id=app_id,
        access_token=access_token,
        states=appstore.editable_version_states,
    )
    print_clr(
        f"Found {colorama.Fore.CYAN}{len(infos)}{colorama.Fore.RESET} editable app infos."
    )

    asset_locales = [
        x for x in os.listdir(app_dir) if os.path.isdir(os.path.join(app_dir, x))
    ]

    for info in infos:
        info_id = info["id"]
        version_state = info["attributes"]["appStoreState"]

        print_clr(
            colorama.Fore.GREEN + "AppInfo ",
            colorama.Fore.BLUE + f"{info_id} ",
            colorama.Fore.CYAN + f"{version_state}",
        )

        localizations = appstore.get_info_localizations(
            info_id=info_id, access_token=access_token
        )

        # create new localizations
        info_locales = [loc["attributes"]["locale"] for loc in localizations]
        new_locales = [x for x in asset_locales if x not in info_locales]
        for locale in new_locales:
            print_locale_status(
                locale, colorama.Fore.LIGHTBLACK_EX, "locale creation not allowed"
            )

        for loc in localizations:
            loc_id = loc["id"]
            loc_attr = loc["attributes"]
            locale = loc_attr["locale"]
            loc_dir = os.path.join(app_dir, locale)

            # Delete removed locales
            if not os.path.isdir(loc_dir):
                print_locale_status(
                    locale, colorama.Fore.LIGHTBLACK_EX, "locale deletion not allowed"
                )
                continue

            # Normalize all attribute values to strings
            for key in appstore.InfoLocalizationAttributes.__annotations__.keys():
                if loc_attr[key] is None:
                    loc_attr[key] = ""

            # Load local data from disk
            asset_loc_data: appstore.InfoLocalizationAttributes = {}
            for key in appstore.InfoLocalizationAttributes.__annotations__.keys():
                path = os.path.join(loc_dir, key + ".txt")
                content = read_txt_file(path)
                if content is not None:
                    asset_loc_data[key] = content  # type: ignore

            # Only need to update if there are differences
            loc_diff_keys = [
                key
                for key, value in asset_loc_data.items()
                if value is not None and value != loc_attr[key]
            ]
            if len(loc_diff_keys) > 0:
                print_locale_status(
                    locale,
                    colorama.Fore.CYAN,
                    f"updating app info {colorama.Fore.CYAN}{colorama.Style.DIM}{loc_diff_keys}",
                )
                appstore.update_info_localization(
                    info_localization_id=loc_id,
                    info_localization_attributes=asset_loc_data,
                    access_token=access_token,
                )
            else:
                print_locale_status(
                    locale, colorama.Fore.CYAN, "no changes in app settings"
                )


def publish(
    access_token: str,
    asset_dir: str,
    app_id: str,
    bundle_id: str,
    platform: Union[appstore.Platform, str],  # pylint: disable=unsubscriptable-object
    version_string: str,
    update_version_string: bool,
    allow_create_version: bool = True,
    allow_create_locale: bool = True,
    allow_delete_locale: bool = True,
):
    """Publish all the app meta data app store, using any editable app versions found.
    If none are found, a new version can be created for the specified target platform."""
    print_clr("Publishing assets from directory: ", colorama.Fore.CYAN + asset_dir)

    # Application directory
    app_dir = os.path.join(asset_dir, bundle_id)
    if not os.path.isdir(app_dir):
        raise FileNotFoundError(
            f"App directory {colorama.Fore.CYAN}{app_dir}{colorama.Fore.RESET} not found. "
        )

    publish_version(
        access_token=access_token,
        app_dir=app_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        platform=platform,
        version_string=version_string,
        update_version_string=update_version_string,
        allow_create_version=allow_create_version,
        allow_create_locale=allow_create_locale,
        allow_delete_locale=allow_delete_locale,
    )
    publish_info(
        access_token=access_token,
        app_dir=app_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        platform=platform,
    )
    print_clr(colorama.Fore.GREEN + "Publish complete")



import colorama
import sys
import os
import logging
import requests
from typing import Union, Optional
from appstore_tools import appstore
from appstore_tools.print_util import print_clr, clr, json_term
from appstore_tools.tqdm_util import tqdm_with_redirect
from .util import (
    write_txt_file,
    write_binary_file,
    fetch_screenshot,
    fetch_preview,
    print_info_status,
    print_version_status,
    print_locale_status,
    print_media_set_status,
    print_media_status,
)


def download_version(
    access_token: str,
    app_dir: str,
    app_id: str,
    bundle_id: str,
    platforms: appstore.PlatformList,
    version_states: appstore.VersionStateList = list(appstore.VersionState),
):
    """Download the app version localized strings and media (screenshots/previews) to the app directory."""
    versions = appstore.get_versions(
        app_id=app_id,
        access_token=access_token,
        platforms=platforms,
        states=version_states,
    )
    if len(versions) == 0:
        message = f"No app version found: {colorama.Fore.CYAN}{platforms}{colorama.Fore.RESET}"
        if version_states != list(appstore.VersionState):
            message += f", {colorama.Fore.CYAN}{version_states}{colorama.Fore.RESET}"
        raise appstore.ResourceNotFoundException(clr(message))

    version = versions[0]
    version_id = version["id"]
    version_state = version["attributes"]["appStoreState"]
    version_platform = version["attributes"]["platform"]

    print_version_status(
        version_state,
        version_platform,
        "downloading app version asset list",
    )

    # Build a full list of needed assets
    localizations = appstore.get_version_localizations(
        version_id=version_id, access_token=access_token
    )
    asset_total = 0
    asset_size_total = 0
    for loc in localizations:
        loc["screenshotSets"] = appstore.get_screenshot_sets(
            localization_id=loc["id"], access_token=access_token
        )
        for screenshot_set in loc["screenshotSets"]:
            screenshots = appstore.get_screenshots(
                screenshot_set_id=screenshot_set["id"], access_token=access_token
            )
            asset_total += len(screenshots)
            asset_size_total += sum(x["attributes"]["fileSize"] for x in screenshots)
            screenshot_set["screenshots"] = screenshots

        loc["previewSets"] = appstore.get_preview_sets(
            localization_id=loc["id"], access_token=access_token
        )
        for preview_set in loc["previewSets"]:
            previews = appstore.get_previews(
                preview_set_id=preview_set["id"], access_token=access_token
            )
            asset_total += len(previews)
            asset_size_total += sum(x["attributes"]["fileSize"] for x in previews)
            preview_set["previews"] = previews

    # Download/Write all the assets
    with tqdm_with_redirect(
        total=asset_size_total, unit="B", unit_scale=True, colour="green", leave=False
    ) as progress_bar:
        for loc in localizations:
            loc_attr = loc["attributes"]
            locale = loc_attr["locale"]
            loc_dir = os.path.join(app_dir, locale)
            screenshots_dir = os.path.join(loc_dir, "screenshots")
            previews_dir = os.path.join(loc_dir, "previews")

            print_locale_status(
                locale, colorama.Fore.CYAN, "downloading version locale"
            )

            # Locale directories
            os.makedirs(name=loc_dir, exist_ok=True)
            os.makedirs(name=screenshots_dir, exist_ok=True)
            os.makedirs(name=previews_dir, exist_ok=True)

            for key in appstore.VersionLocalizationAttributes.__annotations__.keys():
                content = loc_attr[key] if loc_attr[key] is not None else ""
                write_txt_file(
                    path=os.path.join(loc_dir, key + ".txt"),
                    content=content,
                )

            for screenshot_set in loc["screenshotSets"]:
                display_type = screenshot_set["attributes"]["screenshotDisplayType"]
                screenshot_set_dir = os.path.join(screenshots_dir, display_type)

                # Screenshot Set directory
                print_media_set_status(
                    display_type, colorama.Fore.CYAN, "downloading screenshot set"
                )
                os.makedirs(name=screenshot_set_dir, exist_ok=True)

                for screenshot in screenshot_set["screenshots"]:
                    file_name = screenshot["attributes"]["fileName"]
                    file_size = screenshot["attributes"]["fileSize"]
                    file_path = os.path.join(screenshot_set_dir, file_name)

                    response = fetch_screenshot(screenshot)
                    if response is None:
                        print_media_status(
                            file_name, colorama.Fore.RED, "no asset (in processing)"
                        )
                    elif response.ok:
                        print_media_status(
                            file_name, colorama.Fore.CYAN, "writing to disk"
                        )
                        write_binary_file(
                            path=file_path,
                            content=response.content,
                        )
                    else:
                        print_media_status(
                            file_name, colorama.Fore.RED, "download failed"
                        )
                    progress_bar.update(file_size)

            for preview_set in loc["previewSets"]:
                preview_type = preview_set["attributes"]["previewType"]
                preview_set_dir = os.path.join(previews_dir, preview_type)

                # Preview Set directory
                print_media_set_status(
                    preview_type, colorama.Fore.CYAN, "downloading preview set"
                )
                os.makedirs(name=preview_set_dir, exist_ok=True)

                for preview in preview_set["previews"]:
                    file_name = preview["attributes"]["fileName"]
                    file_path = os.path.join(preview_set_dir, file_name)

                    response = fetch_preview(preview)
                    if response is None:
                        print_media_status(
                            file_name,
                            colorama.Fore.RED,
                            "no asset (in processing)",
                        )
                    elif response.ok:
                        print_media_status(
                            file_name, colorama.Fore.CYAN, "writing to disk"
                        )
                        write_binary_file(
                            path=file_path,
                            content=response.content,
                        )
                    else:
                        print_media_status(
                            file_name, colorama.Fore.RED, "download failed"
                        )
                    progress_bar.update(file_size)


def download_info(
    access_token: str,
    app_dir: str,
    app_id: str,
    bundle_id: str,
    version_states: appstore.VersionStateList = list(appstore.VersionState),
):
    """Download the app info to the local app directory."""
    infos = appstore.get_infos(
        app_id=app_id,
        access_token=access_token,
        states=version_states,
    )
    if len(infos) == 0:
        message = f"No app infos found"
        if version_states != list(appstore.VersionState):
            message += f": {colorama.Fore.CYAN}{version_states}{colorama.Fore.RESET}"
        raise appstore.ResourceNotFoundException(clr(message))

    info = infos[0]
    info_id = info["id"]
    info_state = info["attributes"]["appStoreState"]

    print_info_status(info_state, "downloading app info")

    # Info Localizations
    localizations = appstore.get_info_localizations(
        info_id=info_id, access_token=access_token
    )

    for loc in localizations:
        loc_attr = loc["attributes"]
        locale = loc_attr["locale"]
        loc_dir = os.path.join(app_dir, locale)

        print_locale_status(locale, colorama.Fore.CYAN, "downloading info locale")

        # Locale directory
        os.makedirs(name=loc_dir, exist_ok=True)

        for key in appstore.InfoLocalizationAttributes.__annotations__.keys():
            content = loc_attr[key] if loc_attr[key] is not None else ""
            write_txt_file(
                path=os.path.join(loc_dir, key + ".txt"),
                content=content,
            )


def download(
    access_token: str,
    asset_dir: str,
    app_id: str,
    bundle_id: str,
    platforms: appstore.PlatformList,
    version_states: appstore.VersionStateList = list(appstore.VersionState),
    overwrite: bool = False,
):
    """Download all the app meta data to the local app directory."""
    app_dir = os.path.join(asset_dir, bundle_id)

    print_clr(
        f"{colorama.Fore.CYAN}{bundle_id} ",
        f"{colorama.Fore.BLUE}{app_id} ",
        f"-> ",
        f"{colorama.Fore.CYAN}{app_dir}",
    )

    # App
    if os.path.isdir(app_dir) and not overwrite:
        raise FileExistsError(
            f"App directory {colorama.Fore.CYAN}{app_dir}{colorama.Fore.RESET} already exists. "
            + "Specify '--overwrite' if you wish to force downloading to an existing directory."
        )

    download_info(
        access_token=access_token,
        app_dir=app_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        version_states=version_states,
    )
    download_version(
        access_token=access_token,
        app_dir=app_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        platforms=platforms,
        version_states=version_states,
    )
    print_clr(colorama.Fore.GREEN + "Download complete")



import os
import requests
import colorama
from typing import Union
from enum import Enum, auto
from appstore_tools.print_util import print_clr, clr, json_term


def fetch_screenshot(screenshot: dict):
    """Fetches screenshot data. Retuns None if screenshot has no asset."""
    attr = screenshot["attributes"]
    file_ext = os.path.splitext(attr["fileName"])[1]
    if attr["imageAsset"] is None:
        return None

    width = attr["imageAsset"]["width"]
    height = attr["imageAsset"]["height"]
    url_template = attr["imageAsset"]["templateUrl"]

    url = url_template.format(w=width, h=height, f=file_ext[1:])
    return requests.get(url)


def fetch_preview(preview: dict):
    """Fetches screenshot data. Retuns None if screenshot has no asset."""
    attr = preview["attributes"]
    url = attr["videoUrl"]
    if url is None:
        return None
    else:
        return requests.get(url)


def write_binary_file(path: str, content: bytes) -> None:
    with open(file=path, mode="wb") as file:
        file.write(content)


def write_txt_file(path: str, content: str) -> None:
    with open(file=path, mode="w") as file:
        file.write(content)


def read_txt_file(
    path: str,
) -> Union[str, None]:  # pylint: disable=unsubscriptable-object
    try:
        with open(file=path, mode="r") as file:
            return file.read()
    except FileNotFoundError:
        return None


def print_info_status(info_state: str, status: str):
    print_clr(f"{colorama.Fore.CYAN}{info_state}", f" - {status}")


def print_version_status(version_state: str, platform: str, status: str):
    print_clr(
        f"{colorama.Fore.CYAN}{version_state} ",
        f"{colorama.Fore.CYAN + colorama.Style.DIM}{{{platform}}}",
        f" - {status}",
    )


def print_locale_status(locale: str, color: str, status: str):
    print_clr(f"  {color}{locale:5}", f" - {status}")


def print_media_set_status(display_type: str, color: str, status: str):
    print_clr(f"    {color}{display_type}", f" - {status}")


def print_media_status(file_name: str, color: str, status: str):
    print_clr(f"      {color}{file_name}", f" - {status}")



import time
import jwt

APPSTORE_AUDIENCE = "appstoreconnect-v1"
APPSTORE_JWT_ALGO = "ES256"


def create_access_token(issuer_id: str, key_id: str, key: str) -> str:
    """Create an access token for use in the AppStore Connect API."""

    # The token's expiration time, in Unix epoch time; tokens that expire more than
    # 20 minutes in the future are not valid (Ex: 1528408800)
    experation = int(time.time()) + 20 * 60

    # AppStore JWT
    # https://developer.apple.com/documentation/appstoreconnectapi/generating_tokens_for_api_requests
    access_token = jwt.encode(
        {"iss": issuer_id, "exp": experation, "aud": APPSTORE_AUDIENCE},
        key,
        algorithm=APPSTORE_JWT_ALGO,
        headers={"kid": key_id},
    )
    return access_token



import logging
import colorama
import json
import requests
import gzip
from enum import Enum, auto
from typing import Union
from appstore_tools.print_util import clr, json_term
from .util import enum_name
from .exceptions import ResourceNotFoundException

APPSTORE_URI_ROOT = "https://api.appstoreconnect.apple.com/v1"


class FetchMethod(Enum):
    GET = auto()
    POST = auto()
    PATCH = auto()
    DELETE = auto()


def fetch(
    method: Union[FetchMethod, str],  # pylint: disable=unsubscriptable-object
    path: str,
    access_token: str,
    headers: dict = {},
    data=None,
):
    """Fetch a URL resource via the AppStore connect api."""
    headers = {"Authorization": f"Bearer {access_token}", **headers}

    url = APPSTORE_URI_ROOT + path if path.startswith("/") else path

    logging.debug(
        clr(
            f"{colorama.Fore.GREEN}appstore.fetch: {enum_name(method)} {colorama.Fore.MAGENTA}{url}\n",
            f"{colorama.Fore.BLUE}request body:\n",
            json_term(data),
        )
    )

    if not isinstance(method, FetchMethod):
        try:
            method = FetchMethod[method]
        except KeyError:
            raise ValueError(
                f"{method} is not a valid FetchMethod. Options are {list(FetchMethod)}"
            )
    if method == FetchMethod.GET:
        response = requests.get(url=url, headers=headers)
    elif method == FetchMethod.POST:
        headers["Content-Type"] = "application/json"
        response = requests.post(url=url, headers=headers, data=json.dumps(data))
    elif method == FetchMethod.PATCH:
        headers["Content-Type"] = "application/json"
        response = requests.patch(url=url, headers=headers, data=json.dumps(data))
    elif method == FetchMethod.DELETE:
        response = requests.delete(url=url, headers=headers)

    content_type = response.headers["content-type"]

    if content_type == "application/json":
        result = response.json()
        logging.debug(clr(f"{colorama.Fore.BLUE}response body:\n", json_term(result)))
    elif content_type == "application/a-gzip":
        # TODO implement stream decompress
        zipped_data = b""
        for chunk in response.iter_content(1024 * 1024):
            if chunk:
                zipped_data += chunk

        unzipped_data = gzip.decompress(zipped_data)
        result = unzipped_data.decode("utf-8")
    else:
        result = response

    # raise exceptions for easier handling
    if response.status_code == 404:
        raise ResourceNotFoundException(
            f'{method.name} {url} (HttpError {response.status_code})\n{json_term({"request": data, "response":result})}'
        )
    elif not response.ok:
        raise requests.exceptions.HTTPError(
            f'{url} {method.name} (HttpError {response.status_code})\n{json_term({"request": data, "response":result})}'
        )

    return result



import sys
import os
import appstore_tools.appstore as appstore
import appstore_tools.console as console
import appstore_tools.actions as actions


def get_access_token(args):
    try:
        access_token = appstore.create_access_token(
            issuer_id=args.issuer_id, key_id=args.key_id, key=args.key
        )
        return access_token
    except ValueError as error:
        sys.exit(error)


def get_app_id(args, access_token):
    if args.app_id == None:
        args.app_id = appstore.get_app_id(
            bundle_id=args.bundle_id, access_token=access_token
        )
    return args.app_id


def get_bundle_id(args, access_token):
    if args.bundle_id == None:
        args.bundle_id = appstore.get_bundle_id(
            app_id=args.app_id, access_token=access_token
        )
    return args.bundle_id


def list_categories(args):
    access_token = get_access_token(args)
    platforms = console.create_platform_filter_list(args)

    actions.list_categories(
        access_token=access_token,
        platforms=platforms,
        verbosity=args.verbosity,
    )


def list_apps(args):
    actions.list_apps(
        access_token=get_access_token(args),
        verbosity=args.verbosity,
    )


def list_versions(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    platforms = console.create_platform_filter_list(args)
    states = console.create_version_state_filter_list(args)

    actions.list_versions(
        access_token=access_token,
        app_id=app_id,
        platforms=platforms,
        states=states,
        verbosity=args.verbosity,
    )


def list_infos(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    states = console.create_version_state_filter_list(args)

    actions.list_infos(
        access_token=access_token,
        app_id=app_id,
        states=states,
        verbosity=args.verbosity,
    )


def list_screenshots(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    platforms = console.create_platform_filter_list(args)
    states = console.create_version_state_filter_list(args)

    actions.list_screenshots(
        access_token=access_token,
        app_id=app_id,
        platforms=platforms,
        states=states,
        version_limit=args.version_limit,
        verbosity=args.verbosity,
    )


def list_previews(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    platforms = console.create_platform_filter_list(args)
    states = console.create_version_state_filter_list(args)

    actions.list_previews(
        access_token=access_token,
        app_id=app_id,
        platforms=platforms,
        states=states,
        version_limit=args.version_limit,
    )


def download(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    bundle_id = get_bundle_id(args, access_token)

    version_states = (
        [args.version_state]
        if args.version_state is not None
        else list(appstore.VersionState)
    )
    actions.download(
        access_token=access_token,
        asset_dir=args.asset_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        platforms=[args.platform],
        version_states=version_states,
        overwrite=args.overwrite,
    )


def publish(args):
    access_token = get_access_token(args)
    app_id = get_app_id(args, access_token)
    bundle_id = get_bundle_id(args, access_token)

    actions.publish(
        access_token=access_token,
        asset_dir=args.asset_dir,
        app_id=app_id,
        bundle_id=bundle_id,
        platform=args.platform,
        version_string=args.version_string or args.created_version_string,
        update_version_string=args.version_string is not None,
        allow_create_version=not args.no_create_version,
        allow_create_locale=not args.no_create_locale,
        allow_delete_locale=not args.no_delete_locale,
    )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['appstore_tools', 'appstore_tools.actions', 'appstore_tools.appstore']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0',
 'PyJWT>=2.0.1,<3.0.0',
 'Pygments>=2.7.4,<3.0.0',
 'argparse-color-formatter>=1.2.2,<2.0.0',
 'colorama>=0.4.4,<0.5.0',
 'cryptography>=3.3.1,<4.0.0',
 'requests>=2.25.1,<3.0.0',
 'tqdm>=4.56.0,<5.0.0']

entry_points = \
{'console_scripts': ['appstore-tools = appstore_tools:run']}

setup_kwargs = {
    'name': 'appstore-tools',
    'version': '0.1.3',
    'description': 'Tools for the AppStore Connect API.',
    'long_description': '# appstore-tools\n\nTools for the AppStore Connect API.\n\nThis package provides methods to publish, download, and list information about AppStore meta-data (descriptions, keywords, screenshots, previews, etc).  Combined with a deployment workflow (such as github actions), the AppStore meta-data can be tracked and deployed along side the rest of the app\'s source code and assets.\n\n## Install\n\n```zsh\npip install appstore-tools\n```\n\n## Usage\n\n```zsh\nappstore-tools [-h] [--version] action [args]\n```\n\nExamples:\n\n```zsh\n# List all apps under the app store account\nappstore-tools apps\n\n# Download the assets for an app\nappstore-tools download --bundle-id com.example.myapp --asset-dir myassets\n\n# Publish the assets for an app\nappstore-tools publish --bundle-id com.example.myapp --asset-dir myassets\n```\n\n## Usage Config\n\nMost actions will require authentication with the AppStore Connect API, as well as specifying which app to target.\n\nAll these parameters can be passed via command line argument, but for convenience, they (and any others) can also be loaded from a config file.\n\nUse the default config file path of `appstore_tools.config`, or specify another with `--config-file CONFIG_FILE`.\n\n```ini\n; appstore_tools.config\n; sample contents\nissuer-id=xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx\nkey-id=XXXXXXXXXX\nkey-file=/home/me/AppStoreConnect_AuthKey_XXXXXXXXXX.p8\nbundle-id=com.example.myapp\n```\n\n## Code\n\nThe actions provided by the command line can also be accessed by import in a python script.\n\n```python\n# Import the package\nfrom appstore_tools import appstore, actions\n\n# Get the auth credentials\nwith open("AuthKey.p8", "r") as file:\n    key = file.read()\n\nissuer_id="xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"\nkey_id="XXXXXXXXXX"\n\n# Create an access token\naccess_token = appstore.create_access_token(\n    issuer_id=issuer_id, key_id=key_id, key=key\n)\n\n# Call the AppStore connect API\napps = appstore.get_apps(access_token=access_token)\n\n# Or call one of the AppStore-Tools Actions\nactions.list_apps(access_token=access_token)\n\n```\n\n## Asset Directory Structure\n\nThe `download` and `publish` actions look for assets in the following directory structure starting at `--asset-dir ASSET_DIR`. Screenshots and Previews are sorted alphabetically in the store listing.\n\nTo leave an attribute unaffected by the `publish` action, remove the corresponding text file from the tree. An empty text file can be used to set the attribute to an empty string.\n\nLikewise, to leave the screenshots (or previews) unaffected, remove the entire `screenshots` folder. If `screenshots` is present, the `publish` action will add/remove screenshot-display-types and their screenshots to match.\n\n```zsh\n[ASSET_DIR]\n└── com.example.myapp\n    └── en-US\n        ├── description.txt\n        ├── keywords.txt\n        ├── marketingUrl.txt\n        ├── name.txt\n        ├── previews\n        ├── privacyPolicyText.txt\n        ├── privacyPolicyUrl.txt\n        ├── promotionalText.txt\n        ├── screenshots\n        │\xa0\xa0 ├── APP_IPAD_PRO_129\n        │\xa0\xa0 │\xa0\xa0 ├── 10_Image.png\n        │\xa0\xa0 │\xa0\xa0 ├── 20_AnotherImage.png\n        │\xa0\xa0 │\xa0\xa0 ├── 30_MoreImages.png\n        │\xa0\xa0 ├── APP_IPAD_PRO_3GEN_129\n        │\xa0\xa0 │\xa0\xa0 ├── a_is_the_first_letter.png\n        │\xa0\xa0 │\xa0\xa0 ├── b_is_the_second_letter.png\n        │\xa0\xa0 │\xa0\xa0 ├── c_is_the_third_letter.png\n        │\xa0\xa0 ├── APP_IPHONE_55\n        │\xa0\xa0 │\xa0\xa0 ├── image01.png\n        │\xa0\xa0 │\xa0\xa0 ├── image02.png\n        │\xa0\xa0 │\xa0\xa0 ├── image03.png\n        │\xa0\xa0 └── APP_IPHONE_65\n        │\xa0\xa0 │\xa0\xa0 ├── image01.png\n        │\xa0\xa0 │\xa0\xa0 ├── image02.png\n        │\xa0\xa0 │\xa0\xa0 ├── image03.png\n        ├── subtitle.txt\n        ├── supportUrl.txt\n        └── whatsNew.txt\n```\n\n## Source\n\nClone the source code\n\n```zsh\ngit clone https://github.com/bennord/appstore-tools.git\n```\n\nInstall dependencies\n\n```zsh\npoetry install\n```\n\nRun from within project environment\n\n```zsh\npoetry shell\nappstore-tools --version\n```\n',
    'author': 'Ben Nordstrom',
    'author_email': 'bennord@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bennord/appstore-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
